(* generated by Ott 0.21.2 from: lang2.ott *)

Require Import Arith.
Require Import Bool.
Require Import List.


Require Import Omega ClassicalChoice basis (* LibMap *). 


Set Implicit Arguments.

Definition nf_update B (r: nat -> B) (x : nat) (b: B) (y : nat) : B :=
  if Peano_dec.eq_nat_dec x y then b else r y.

Definition ffun B :=
  { f: nat -> option B & { n | forall m (LEQ: n <= m), f m = None} }.

Program Definition empty_map B : ffun B := 
  existT _ (fun y => None) (existT _ 0 _).

Program Definition ff_update B (r: ffun B) (x : nat) (b: B) : ffun B :=
  existT _ (fun y => if Peano_dec.eq_nat_dec x y then Some b else projT1 r y)
   (existT _ (max (projT1 (projT2 r)) (S x)) _).
Next Obligation.
  destruct r as (r & n & H); simpl in *.
  destruct (Peano_dec.eq_nat_dec x m); subst.
  eapply Max.max_lub_r in LEQ; omega. 
  eauto using Max.max_lub_l.
Qed.

 
Definition termvar := nat.
Lemma eq_termvar: forall (x y : termvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_termvar : ott_coq_equality.

Inductive exp : Set := 
 | Evar (x:termvar)
 | Eplus (E1:exp) (E2:exp).

Inductive bexp : Set := 
 | Beq (E1:exp) (E2:exp)
 | Bless (E1:exp) (E2:exp).

Inductive basiccmd : Set := 
 | BCassign (x:termvar) (E:exp)
 | BCassume (B:bexp).

Inductive cmd : Set := 
 | Cskip : cmd
 | Cbasic (c:basiccmd)
 | Cseq (C1:cmd) (C2:cmd)
 | Cpar (C1:cmd) (C2:cmd)
 | Cchoice (C1:cmd) (C2:cmd)
 | Cloopiter (C1:cmd) (C2:cmd) (n:nat).
(** definitions *)

(** funs edenot_funs *)
Fixpoint edenot (x1:exp) (x2:(nat -> nat)) : nat:=
  match x1,x2 with
  | (Evar x) , ss =>  ( ss   x ) 
  | (Eplus E1 E2) , ss =>  (  (edenot E1 ss )   +   (edenot E2 ss )  ) 
end.

(** definitions *)

(** funs bdenot_funs *)
Fixpoint bdenot (x1:bexp) (x2:(nat -> nat)) : Prop:=
  match x1,x2 with
  | (Beq E1 E2) , ss =>   (edenot E1 ss )   =   (edenot E2 ss )  
  | (Bless E1 E2) , ss =>   (edenot E1 ss )   <   (edenot E2 ss )  
end.

(** definitions *)

(** funs cdenot_funs *)
Fixpoint cdenot (x1:basiccmd) (x2:(nat -> nat)) (x3:(nat -> nat)) : Prop:=
  match x1,x2,x3 with
  | (BCassign x E) , ss , ss' =>  ss'  =   (nf_update  ss   x    (edenot E ss )  )  
  | (BCassume  B ) , ss , ss' =>   ss'  =  ss   /\   (bdenot B ss )  
end.

(** definitions *)

(** funs ncons_funs *)
Fixpoint ncons (x1:nat) (x2:nat) : list nat:=
  match x1,x2 with
  | n ,  0  =>  nil 
  | n ,   (S  m )   =>  ( n ::  (ncons n m )  ) 
end.

(** definitions *)

(** funs merge_funs *)
Fixpoint merge (x1:list nat) (x2:list nat) : list nat:=
  match x1,x2 with
  |  nil  , P' => P'
  |  ( p :: P )  ,  nil  =>  ( p :: P ) 
  |  ( p :: P )  ,  ( p' :: P' )  =>  (   ( p  +  p' )   ::  (  (merge P P' )  )  ) 
end.

(** definitions *)

(** funs size_funs *)
Fixpoint size (x1:cmd) (x2:list nat) : list nat:=
  match x1,x2 with
  | Cskip , P =>  (ncons  0   (length  P )  ) 
  | (Cbasic c) , P =>  (ncons  1   (length  P )  ) 
  | (Cseq C1 C2) , P =>  (merge  (merge  (size C1 P )   (size C2 P )  )   (ncons  1   (length  P )  )  ) 
  | (Cloopiter C1 C2 n) ,  nil  =>  nil 
  | (Cloopiter C1 C2 n) ,  ( p :: P )  =>  (  1  ::  (  (if Compare_dec.lt_dec  n   p  then   (merge  (merge  (size C1 P )   (size C2 P )  )   (ncons  1   (length  P )  )  )   else   (size C1 P )  )  )  ) 
  | (Cpar C1 C2) , P =>  (merge  (merge  (size C1 P )   (size C2 P )  )   (ncons  1   (length  P )  )  ) 
  | (Cchoice C1 C2) , P =>  (merge  (merge  (size C1 P )   (size C2 P )  )   (ncons  1   (length  P )  )  ) 
end.

(** definitions *)

(** funs loop_depth_funs *)
Fixpoint loop_depth (x1:cmd) : nat:=
  match x1 with
  | Cskip =>  0 
  | (Cbasic c) =>  0 
  | (Cloopiter C1 C2 n) =>  (  1   +   (max   (loop_depth C1 )     (loop_depth C2 )  )  ) 
  | (Cseq C1 C2) =>  (max   (loop_depth C1 )     (loop_depth C2 )  ) 
  | (Cpar C1 C2) =>  (max   (loop_depth C1 )     (loop_depth C2 )  ) 
  | (Cchoice C1 C2) =>  (max   (loop_depth C1 )     (loop_depth C2 )  ) 
end.

(** definitions *)

(** funs user_cmd_funs *)
Fixpoint user_cmd (x1:cmd) : Prop:=
  match x1 with
  |  Cskip  =>  True 
  |  (Cbasic c)  =>  True 
  |  (Cloopiter Cskip C2  0 )  =>  (user_cmd C2 ) 
  |  (Cloopiter C1 C2 p)  =>  False 
  |  (Cseq C1 C2)  =>   (user_cmd C1 )   /\   (user_cmd C2 )  
  |  (Cpar C1 C2)  =>   (user_cmd C1 )   /\   (user_cmd C2 )  
  |  (Cchoice C1 C2)  =>   (user_cmd C1 )   /\   (user_cmd C2 )  
end.

(** definitions *)

(** funs user_conf_funs *)
Fixpoint user_conf (x1:cmd) (x2:list nat) : Prop:=
  match x1,x2 with
  | Cskip ,  nil  =>  False 
  | Cskip ,  ( p :: P )  =>  True 
  | (Cbasic c) ,  nil  =>  False 
  | (Cbasic c) ,  ( p :: P )  =>  True 
  | (Cloopiter Cskip C2  0 ) ,  ( p :: P )  =>  (user_conf C2 P ) 
  | (Cloopiter C1 C2 p0) , P =>  False 
  | (Cseq C1 C2) , P =>   (user_conf C1 P )   /\   (user_conf C2 P )  
  | (Cpar C1 C2) , P =>   (user_conf C1 P )   /\   (user_conf C2 P )  
  | (Cchoice C1 C2) , P =>   (user_conf C1 P )   /\   (user_conf C2 P )  
end.

(** definitions *)

(** funs wf_conf_funs *)
Fixpoint wf_conf (x1:cmd) (x2:list nat) : Prop:=
  match x1,x2 with
  | Cskip ,  nil  =>  False 
  | Cskip ,  ( p :: P )  =>  True 
  | (Cbasic c) ,  nil  =>  False 
  | (Cbasic c) ,  ( p :: P )  =>  True 
  | (Cloopiter C1 C2 n) ,  nil  =>  False 
  | (Cloopiter C1 C2 n) ,  ( p :: P )  =>    (wf_conf C1 P )   /\   (user_conf C2 P )    /\    (loop_depth C1 )   <=   (loop_depth C2 )   
  | (Cseq C1 C2) , P =>   (wf_conf C1 P )   /\   (user_conf C2 P )  
  | (Cpar C1 C2) , P =>   (wf_conf C1 P )   /\   (wf_conf C2 P )  
  | (Cchoice C1 C2) , P =>   (user_conf C1 P )   /\   (user_conf C2 P )  
end.

(** definitions *)

(* defns OpSemJudg *)
Inductive step : cmd -> (nat -> nat) -> list nat -> cmd -> (nat -> nat) -> list nat -> Prop :=    (* defn step *)
 | Sbasic : forall (c:basiccmd) (ss:(nat -> nat)) (P:list nat) (ss':(nat -> nat)),
      (cdenot c ss ss' )  ->
     step (Cbasic c) ss P Cskip ss'  (merge P  (ncons  1   (length  P )  )  ) 
 | SseqI : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat) (C1':cmd) (ss':(nat -> nat)) (P':list nat),
     step C1 ss P C1' ss' P' ->
     step (Cseq C1 C2) ss P (Cseq C1' C2) ss' P'
 | SseqII : forall (C2:cmd) (ss:(nat -> nat)) (P:list nat),
     step (Cseq Cskip C2) ss P C2 ss P
 | SparEnd : forall (ss:(nat -> nat)) (P:list nat),
     step (Cpar Cskip Cskip) ss P Cskip ss P
 | SparL : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat) (C1':cmd) (ss':(nat -> nat)) (P':list nat),
     step C1 ss P C1' ss' P' ->
     step (Cpar C1 C2) ss P (Cpar C1' C2) ss' P'
 | SparR : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat) (C2':cmd) (ss':(nat -> nat)) (P':list nat),
     step C2 ss P C2' ss' P' ->
     step (Cpar C1 C2) ss P (Cpar C1 C2') ss' P'
 | SnondetL : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat),
     step (Cchoice C1 C2) ss P C1 ss P
 | SnondetR : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat),
     step (Cchoice C1 C2) ss P C2 ss P
 | SloopI : forall (C:cmd) (n:nat) (ss:(nat -> nat)) (p:nat) (P:list nat),
     step (Cloopiter Cskip C n) ss  ( p :: P )  Cskip ss  ( p :: P ) 
 | SloopII : forall (C:cmd) (n:nat) (ss:(nat -> nat)) (p:nat) (P:list nat),
     step (Cloopiter Cskip C n) ss  ( p :: P )  (Cloopiter C C p) ss  ( p :: P ) 
 | SloopIII : forall (C1 C2:cmd) (n:nat) (ss:(nat -> nat)) (p:nat) (P:list nat) (C1':cmd) (ss':(nat -> nat)) (P':list nat),
     step C1 ss P C1' ss' P' ->
     step (Cloopiter C1 C2 n) ss  ( p :: P )  (Cloopiter C1' C2 n) ss'  ( p :: P' ) .
(** definitions *)

(* defns aborts_def *)
Inductive aborts : cmd -> (nat -> nat) -> list nat -> Prop :=    (* defn aborts *)
 | aborts_seqI : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat),
     aborts C1 ss P ->
     aborts (Cseq C1 C2) ss P
 | aborts_parI : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat),
     aborts C1 ss P ->
     aborts (Cpar C1 C2) ss P
 | aborts_parII : forall (C1 C2:cmd) (ss:(nat -> nat)) (P:list nat),
     aborts C2 ss P ->
     aborts (Cpar C1 C2) ss P
 | aborts_loopI : forall (C1 C2:cmd) (n:nat) (ss:(nat -> nat)) (p:nat) (P:list nat),
     aborts C1 ss P ->
     aborts (Cloopiter C1 C2 n) ss  ( p :: P ) 
 | aborts_loopII : forall (C:cmd) (n:nat) (ss:(nat -> nat)) (p:nat) (P:list nat),
      p  <=  n  ->
     aborts (Cloopiter Cskip C n) ss  ( p :: P ) .


