Require Import List Vbase basis lang2.
Require Import Wellfounded NPeano.

Set Implicit Arguments.

Lemma loop_depth_inv :
  forall C ss p C' ss' p',
    step C ss p C' ss' p' ->
    loop_depth C' <= loop_depth C.
Proof.
  induction 1; ins; auto using Max.max_lub, Nat.max_le_compat with arith. 
Qed.

Lemma user_conf_wf_conf :
  forall C P, user_conf C P -> wf_conf C P.
Proof.
  induction C; ins; desf; ins; intuition. 
  clear - H; induction C2; ins; desf; ins; auto; omega.
Qed.

Lemma user_confE : 
  forall C P, user_conf C P <-> user_cmd C /\ loop_depth C < length P.
Proof.
  induction C; ins; desf; ins; rewrite ?IHC1, ?IHC2, ?IHC;
  repeat split; ins; desf; rewrite ?Nat.max_lub_lt_iff in *; omega.
Qed.

Hint Resolve user_conf_wf_conf.

Inductive allp A (X: A -> A -> Prop) : list A -> list A -> Prop :=
  | allp_nil : allp X nil nil
  | allp_cons a b (HD : X a b) l l' (TL : allp X l l') :
      allp X (cons a l) (cons b l').

Lemma allp_le_refl P : allp le P P.
Proof. by induction P; vauto. Qed.

Lemma allp_le_pone P : allp le P (merge P (ncons 1 (length P))).
Proof. by induction P; ins; constructor; auto; omega. Qed.
Hint Resolve allp_le_refl allp_le_pone.

Lemma allp_length A (X: A -> A -> Prop) l l' : 
  allp X l l' -> length l = length l'.
Proof. induction 1; ins; omega. Qed.

Lemma step_allp :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    allp le P P'.
Proof.
  induction 1; ins; desf; intuition; vauto.
Qed.

Hint Resolve step_allp.

Lemma wf_inv :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    wf_conf C P ->
    wf_conf C' P .
Proof.
  induction 1; ins; desf; intuition.
  apply loop_depth_inv in H; omega.
Qed.

Lemma user_mon :
  forall C P P',
    user_conf C P ->
    allp le P P' ->
    user_conf C P'.
Proof.
  induction C; ins; desf; eauto; inv H0; eauto; ins; omega. 
Qed.

Lemma wf_mon :
  forall C P P',
    wf_conf C P ->
    allp le P P' ->
    wf_conf C P'.
Proof.
  induction C; ins; desf; try inv H0; intuition; eauto using user_mon; 
    ins; omega.
Qed.

Lemma length_inv :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    length P = length P'.
Proof.
  induction 1; ins; try omega; induction P; ins; omega.
Qed.

Lemma length_ncons n m : length (ncons n m) = m.
Proof. induction m; ins; eauto. Qed.

Lemma length_merge :
  forall l l', length (merge l l') = max (length l) (length l').
Proof. induction l, l'; ins; eauto. Qed.

Lemma length_size : forall C P, length (size C P) = length P.
Proof.
  induction C; ins; desf;
    repeat (rewrite ?map_length, ?length_ncons, ?length_merge, 
            ?IHC1, ?IHC2, ?Nat.max_id, ?Max.max_0_r; ins).
Qed.

Ltac rewrite_length :=
  repeat (rewrite ?map_length, ?length_ncons, ?length_merge, 
                  ?length_size, ?Nat.max_id; ins).

Lemma mergeC : forall l m, merge l m = merge m l.
Proof. induction l, m; ins; f_equal; auto with arith. Qed.

Inductive lexlt : list nat -> list nat -> Prop :=
  | lexlt_eq a l l' (TL : lexlt l l') : lexlt (cons a l) (cons a l')
  | lexlt_lt a b (HD : a < b) l l' (LEQ : length l <= length l') : 
      lexlt (cons a l) (cons b l').

Lemma lexlt_length l l' : lexlt l l' -> length l <= length l'.
Proof. induction 1; ins; omega. Qed.

Definition lexle l l' := l = l' \/ lexlt l l'.

Lemma lexle_refl : forall x, lexle x x.
Proof. vauto. Qed.

Hint Resolve lexle_refl.

(*
Lemma lexlt_inc l l': lexlt l l' -> lexlt (inc l) (inc l').
Proof. induction 1; constructor; ins; rewrite ?length_inc; omega. Qed.

Lemma lexle_inc l l': lexle l l' -> lexle (inc l) (inc l').
Proof. unfold lexle; ins; desf; eauto using lexlt_inc. Qed.

Hint Resolve lexlt_inc lexle_inc.
*)

Lemma lexltle_merge : 
  forall l l', lexlt l l' -> 
  forall m m', lexle m m' ->
  lexlt (merge l m) (merge l' m').
Proof.
  unfold lexle; induction 1; ins; desf;
    try match goal with H : lexlt _ _ |- _ => inv H end;
    constructor; eauto using lexlt, Plus.plus_lt_compat, Plus.plus_lt_compat_r, Plus.plus_lt_compat_l;
    rewrite ?length_merge; try apply lexlt_length in H; apply Nat.max_le_compat; ins; omega.
Qed.

Lemma lexlelt_merge : 
  forall l l', lexle l l' -> 
  forall m m', lexlt m m' ->
  lexlt (merge l m) (merge l' m').
Proof.
  ins; rewrite (mergeC l), (mergeC l'); eauto using lexltle_merge.
Qed.

Lemma lexle_merge : 
  forall l l', lexle l l' -> 
  forall m m', lexle m m' ->
  lexle (merge l m) (merge l' m').
Proof.
  unfold lexle; ins; desf; vauto; right;
  try solve [apply lexltle_merge; vauto | apply lexlelt_merge; vauto].
Qed.

Hint Resolve lexle_merge.

Lemma step_incrP :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    lexle P P'.
Proof.
  induction 1; ins; desf; vauto; unfold lexle in *; desf; eauto using lexlt.
  destruct P; ins; vauto; right; constructor; rewrite_length; omega.
Qed.

Lemma lexle_cons_eq x l l' : lexle l l' -> lexle (x :: l) (x :: l').
Proof. unfold lexle; ins; desf; vauto. Qed.

Fixpoint take A n (l: list A) := 
  match n, l with 
    | 0, _ => nil
    | _, nil => nil
    | S n, x :: l => x :: take n l
  end.

(*Lemma take_inc : forall n l, take n (inc l) = inc (take n l). 
Proof. induction n, l; ins; f_equal; auto. Qed.*)

Lemma take_merge : forall n l m, take n (merge l m) = merge (take n l) (take n m). 
Proof. induction n, l, m; ins; f_equal; auto. Qed.

Lemma take_map : forall A B (f: A -> B) n l, take n (map f l) = map f (take n l).
Proof. induction n, l; ins; f_equal; eauto. Qed.

(*
Lemma size_exp : 
  forall C P, wf_conf C P -> exists n l, size C P = n :: l.
Proof.
  induction C; ins; desf; eauto;
   try edestruct IHC1 as (? & ? & ->); eauto;
   try edestruct IHC2 as (? & ? & ->); ins; eauto;
   try (by destruct P; ins; eauto; inv H).
Qed.
*)

Lemma lengtheq_map0 :
  forall A (l l' : list A),
    length l = length l' -> forall B (b: B),
    map (fun _ => b) l = map (fun _ => b) l'.
Proof.
  induction l, l'; ins; f_equal; eauto.
Qed.  

Lemma size2_take_eqP :
  forall C P P' n,
    length P = length P' ->
    take (n - 1) P = take (n - 1) P' ->
    take n (size C P') = take n (size C P).
Proof.
  induction C; ins;
    repeat first [rewrite take_map | rewrite take_merge];
    try (by rewrite (lengtheq_map0 _ _ H));
    try by do 2 f_equal; eauto.

  destruct n0; ins; desf; rewrite ?H; ins;
    f_equal; destruct n0; try done;
    repeat first [rewrite take_map | rewrite take_merge];
    simpl in H0; desf; replace n0 with (S n0 - 1) in H0; try omega;
    rewrite ?(IHC1 _ _ _ H), ?(IHC2 _ _ _ H); try done.
Qed.


Lemma lexlt_inc_r l l' m :
  length l <> 0 ->
  m <> 0 ->
  lexle l l' ->
  lexlt l (merge l' (ncons 1 m)).
Proof.
  destruct 3; desf. 
  induction l', m; ins; constructor; rewrite_length; auto with arith; omega.
  induction H1, m; ins; constructor; rewrite_length; auto with arith; try omega;
  rewrite Nat.max_le_iff; auto using lexlt_length.
Qed.

Lemma merge_map0 :
  forall l m,
    length l <= length m -> 
    merge (map (fun _ : nat => 0) l) m = m.
Proof.
  induction l, m; ins; rewrite ?IHl; ins; omega.
Qed.

Lemma wf_conf_length_nonzero : 
  forall C P, wf_conf C P -> length P <> 0.
Proof.
  induction C; ins; desf; eauto.
Qed.

Hint Resolve wf_conf_length_nonzero.

Lemma lexle_merge_l :
  forall l l', length l = length l' -> lexle l (merge l l').
Proof.
  induction l, l'; ins.
  destruct n; desf.
    by rewrite NPeano.Nat.add_0_r; apply lexle_cons_eq, IHl.
  right; constructor; rewrite ?length_merge; auto with arith; omega.
Qed.

Lemma lexle_merge_r :
  forall l l', length l = length l' -> lexle l' (merge l l').
Proof.
  by ins; rewrite mergeC; apply lexle_merge_l.
Qed.

Hint Constructors aborts.

Lemma metric_inv1 :
  forall C ss P C' ss' P'
    (S: step C ss P C' ss' P')
    (NA: ~ aborts C ss P)
    (EQ: P = P')
    (WF: wf_conf C P), 
  lexlt (size C' P') (size C P).
Proof.
  induction 1; ins; intuition; vauto;
    try (by ins; eapply lexltle_merge; [eapply lexltle_merge|]; eauto);
    try (by ins; eapply lexltle_merge; [eapply lexlelt_merge|]; eauto);
    try (by rewrite ?merge_map0; rewrite ?map_length, ?length_size; ins;
         apply lexlt_inc_r; rewrite ?length_size; eauto; desf). 

  by exfalso; destruct P; ins; desf; omega.

  by apply lexlt_inc_r, lexle_merge_r; rewrite_length; desf.
  by apply lexlt_inc_r, lexle_merge_r; rewrite_length; desf.
  by apply lexlt_inc_r, lexle_merge_l; rewrite_length; eauto.
  by apply lexlt_inc_r, lexle_merge_r; rewrite_length; eauto.


  by constructor; ins; desf; rewrite_length. 

  constructor; desf; try (by exfalso; omega).
    by apply lexlt_inc_r, lexle_merge_r; rewrite_length; desf.
  by exfalso; apply NA, aborts_loopII; omega. 

  by constructor; desf; eauto using lexltle_merge.
Qed.

Lemma metric_inv2 :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    P <> P' ->
    wf_conf C P -> 
    exists n, take (n - 1) P = take (n - 1) P' 
      /\ lexlt (take n (size C' P')) (take n (size C P)).
Proof.
  induction 1; ins; intuition; desf.

  by exists 1,; vauto.

  exists n,; vauto; rewrite !take_merge, (length_inv H).
  by apply lexltle_merge; [apply lexltle_merge|]; ins;
     left; apply size2_take_eqP; eauto using length_inv.

  exists n,; vauto; rewrite !take_merge, (length_inv H).
  by apply lexltle_merge; [apply lexltle_merge|]; ins;
     left; apply size2_take_eqP; eauto using length_inv.

  exists n,; vauto; rewrite !take_merge, (length_inv H).
  by apply lexltle_merge; [apply lexlelt_merge|]; ins;
     left; apply size2_take_eqP; eauto using length_inv.

  destruct IHstep as (k & X & Y); ins; desf; eauto.
  exists (S k),; ins; vauto.
    by clear - X; destruct k as [|[]]; ins; desf; congruence.
  constructor; rewrite !take_merge, (length_inv H).
  by apply lexltle_merge; [apply lexltle_merge|]; ins;
     left; apply size2_take_eqP; eauto using length_inv.

  destruct IHstep as (k & X & Y); ins; desf; eauto.
  exists (S k),; ins; vauto.
  by clear - X; destruct k as [|[]]; ins; desf; congruence.
Qed.

Lemma lexlt_takeE :
  forall n l m,
    lexlt (take n l) (take n m) ->
    length l <= length m ->
    lexlt l m.
Proof.
  induction n; ins; desf; inv H; ins; constructor; ins; [eapply IHn; ins|]; omega.
Qed.

Lemma metric_inv :
  forall C ss P C' ss' P',
    step C ss P C' ss' P' ->
    ~ aborts C ss P ->
    wf_conf C P -> 
    lexlt (size C' P') (size C P).
Proof.
  ins.
  destruct (list_eq_dec NPeano.Nat.eq_dec P P'); 
    [by apply metric_inv1 in H; eauto|].
  generalize H; intro M.
  apply metric_inv2 in H; desf; eauto using lexlt_takeE. 
  eapply lexlt_takeE; eauto.
  by rewrite !length_size, (length_inv M). 
Qed.

Lemma wf_lexlt_aux : 
  forall n, well_founded (fun x y => length y <= n /\ <<R:lexlt x y>>).
Proof.
  induction n.
    by constructor; ins; desf; inv R; ins; exfalso; omega.
  intros [|h t]; [by constructor; ins; desf; inv R|].
  assert (WF := wf_lexprod _ _ _ _ Wf_nat.lt_wf (fun _ => IHn) (existT _ h t)).
  clear IHn.
  remember (existT _ h t) as r; revert h t Heqr.
  induction WF; ins; desf.
  constructor; ins; desf.
  inv R; eapply H0; vauto; eauto.
  constructor 2; ins; unfold NW; eauto using le_S_n.
Qed.

Lemma wf_lexlt : well_founded lexlt.
Proof.
  red; ins.
  generalize (Le.le_refl (length a)); generalize (length a) at 2; intro n.
  induction a as [? IH] using (well_founded_ind (wf_lexlt_aux n)).
  intro LT; constructor; intros ? X; apply IH; vauto.
  eapply Le.le_trans, LT.
  clear -X; induction X; ins; omega. 
Qed.

